#!/usr/bin/env python3
# Usage {{{1
"""
Avendesora Password Generator

Generates passwords and pass phrases based on stored account information.

usage:
    avendesora [options]
    avendesora [options] <account> [<secret>]

options:
    -a, --all               Output all account information.
    -b, --browse            Open account in the default browser ({defbrowser}).
    -B <choice>, --browser <choice>
                            Open account in specified browser.  Choose from
                            {browsers}.
    -c, --clipboard         Write output to clipboard rather than stdout.
    -f <text>, --find <text>
                            List any account that contains the given string in 
                            its ID or aliases.
    -g <GPG ID>, --gpgid <GPG ID>
                            Use this GPG ID when creating any missing files.
    -H <plain_text>, --hide <plain_text>
                            Encode plain_text using base64 as a way of hiding 
                            (but not encrypting) its value.
    -I, --init              Create an initial accounts file
    -R <coded_text>, --reveal <coded_text>
                            Decode coded_text using base64.
    -s <text>, --search <text>
                            List any account that contains the given string in 
                            its ID or in any field value that is not encrypted
                            or concealed.
    --stdout                Write output to the standard output without any
                            annotation or protections.
"""

# License {{{1
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see http://www.gnu.org/licenses/.


# Imports {{{1
from .config import read_config, get_setting
from .generator import PasswordGenerator
from .gpg import GnuPG
from .preferences import SETTINGS_DIR
from .secrets import Hidden
from .writer import get_writer
from inform import Inform, Error, output, terminate, debug
from shlib import to_path
import docopt
import sys

# Utilities {{{1
def print_search_results(search_term, search_func):
    to_print = []
    for acct in search_func(search_term):
        aliases = getattr(acct, 'aliases', [])

        aliases = ' (%s)' % (', '.join(aliases)) if aliases else ''
        to_print += [acct.get_name() + aliases]
    output(search_term + ':')
    output('    ' + ('\n    '.join(sorted(to_print))))

# Main {{{1
def main():
    # read config file
    read_config()
    browsers = get_setting('browsers')
    browsers = ', '.join(
        ['%s (%s)' % (k, browsers[k].split()[0]) for k in sorted(browsers)]
    )

    # read command line
    cmdline = docopt.docopt(
        __doc__.format(
            defbrowser=get_setting('default_browser'), browsers=browsers
        )
    )

    # start logging
    def teardown():
        inform.disconnect()
        gpg.close()
    gpg = GnuPG(gpg_id=cmdline['--gpgid'])
    inform = Inform(
        logfile=gpg.open(to_path(SETTINGS_DIR, get_setting('log_file'))),
        termination_callback=teardown
    )

    # process hide/reveal commands
    if cmdline['--hide']:
        debug(cmdline['--hide'])
        output(Hidden.hide(cmdline['--hide']))
        terminate()
    if cmdline['--reveal']:
        output(Hidden.reveal(cmdline['--reveal']))
        terminate()

    # run the generator
    try:
        generator = PasswordGenerator(
            gpg_id = cmdline['--gpgid'],
            init = cmdline['--init']
        )

        # search for accounts that match search criteria
        if cmdline['--find']:
            print_search_results(cmdline['--find'], generator.find_accounts)
            terminate()

        if cmdline['--search']:
            print_search_results(cmdline['--search'], generator.search_accounts)
            terminate()

        # determine the account and output specified information
        account_name = cmdline['<account>']
        writer = get_writer(
            cmdline['<account>'], cmdline['--clipboard'], cmdline['--stdout']
        )
        if account_name:
            account = generator.get_account(account_name)
            if cmdline['--browse'] or cmdline['--browser']:
                account.open_browser(cmdline['--browser'])
            else:
                if cmdline['--all']:
                    account.write_summary()
                if cmdline['<secret>'] or not cmdline['--all']:
                    writer.display_field(account, cmdline['<secret>'])
        else:
            # complain if account is required
            if cmdline['--browse'] or cmdline['--browser']:
                fatal('account missing (required when starting browser).')

            # use discovery to determine account
            account_name, script = generator.discover_account()
            account = generator.get_account(account_name)
            writer.run_script(account, script)

    except KeyboardInterrupt:
        output('Terminated by user.')
    except Error as err:
        err.terminate()

    terminate()



from .charsets import (
    exclude, LOWERCASE, UPPERCASE, LETTERS, DIGITS, ALPHANUMERIC,
    HEXDIGITS, PUNCTUATION, WHITESPACE, PRINTABLE, DISTINGUISHABLE
)
from .account import Account
from .secrets import (
    Password, Passphrase, PIN, Question, MixedPassword, BirthDate,
    Hidden, GPG
)
from .recognizers import (
    RecognizeAll, RecognizeAny, RecognizeTitle, RecognizeURL, RecognizeCWD,
    RecognizeHost, RecognizeUser, RecognizeEnvVar
)
__version__ = '0.1.12'



# Utilities

# License {{{1
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see http://www.gnu.org/licenses/.

# Imports {{{1
from .config import get_setting
from .dictionary import DICTIONARY
from .preferences import (
    SECRETS_MD5, CHARSETS_MD5, SETTINGS_DIR, CONFIG_FILENAME
)
from shlib import Run, to_path
from inform import codicil, error, os_error, warn
from textwrap import dedent, wrap
import hashlib

# gethostname {{{1
# returns short version of the hostname (the hostname without any domain name)
import socket
def gethostname():
    return socket.gethostname().split('.')[0]

# getusername {{{1
import getpass
def getusername():
    return getpass.getuser()

# generate_random_string {{{1
def generate_random_string(length=64):
    # Generate a random long string to act as the default password

    from string import ascii_letters, digits, punctuation
    import random
    rand = random.SystemRandom()

    # Create alphabet from letters, digits, and punctuation, but 
    # replace double quote with a space so password can be safely 
    # represented as a double-quoted string.
    alphabet = (ascii_letters + digits + punctuation).replace('"', ' ')

    password = ''
    for i in range(length):
        password += rand.choice(alphabet)
    return password


# validate_componenets {{{1
def validate_componenets():
    # Check that dictionary has not changed.
    # If the master password file exists, then self.data['dict_hash'] will 
    # exist, and we will compare the current hash for the dictionary 
    # against that stored in the master password file, otherwise we will 
    # compare against the one present when the program was configured.
    DICTIONARY.validate(get_setting('dict_hash'))

    # Check that secrets.py and charset.py have not changed
    for each, md5 in [
        ('secrets', SECRETS_MD5),
        ('charsets', CHARSETS_MD5)
    ]:
        src_dir = to_path(__file__).parent
        path = to_path(src_dir, each + '.py')
        try:
            contents = path.read_text()
        except IOError as err:
            path = to_path(src_dir, '..', each + '.py')
            try:
                contents = path.read_text()
            except IOError as err:
                error('%s: %s.' % (err.filename, err.strerror))
        hash = hashlib.md5(contents.encode('utf-8')).hexdigest()
        # Check that file has not changed.
        # If the master password file exists, then self.data['%s_hash'] 
        # will exist, and we will compare the current hash for the file 
        # against that stored in the master password file, otherwise we 
        # will compare against the one present when the program was 
        # configured.
        if hash != get_setting('%s_hash' % each, md5):
            warn("file contents have changed.", culprit=path)
            codicil(
                *wrap(dedent("""\
                    This results in passwords that are inconsistent with those
                    created in the past.  Change {settings}/{config} to contain
                    "{kind}_hash = '{hash}'".  Then use 'avendesora
                    changed' to assure that nothing has changed.
                """.format(
                    kind=each, settings=SETTINGS_DIR, config=CONFIG_FILENAME,
                    hash=hash
                ))),
                sep = '\n'
            )



import pytest
import subprocess
import os

# set various environment variables so avendesora uses local gpg key and config
# directory rather than the users.

os.environ['HOME'] = 'home'

def test_mybank():
    result = subprocess.check_output('avendesora --stdout mybank'.split())
    assert result == b'NNPx2dHiR7mo\n'

def test_mb():
    result = subprocess.check_output('avendesora --stdout mb'.split())
    assert result == b'NNPx2dHiR7mo\n'

def test_mb_checking():
    result = subprocess.check_output('avendesora --stdout mb accounts.checking'.split())
    assert result == b'12345678\n'

def test_mb_savings():
    result = subprocess.check_output('avendesora --stdout mb accounts[savings]'.split())
    assert result == b'23456789\n'

def test_alertscc():
    result = subprocess.check_output('avendesora --stdout alertscc'.split())
    assert result == b'UnmvA52NYMha\n'

def test_scc():
    result = subprocess.check_output('avendesora --stdout scc'.split())
    assert result == b'UnmvA52NYMha\n'

def test_scc_account():
    result = subprocess.check_output('avendesora --stdout scc account'.split())
    assert result == b'123456-7890\n'

def test_scc_birthdate():
    result = subprocess.check_output('avendesora --stdout scc birthdate'.split())
    assert result == b'1969-12-28\n'

def test_scc_email():
    result = subprocess.check_output('avendesora --stdout scc email'.split())
    assert result == b'pizzaman@pizza.com\n'

def test_scc_q0():
    result = subprocess.check_output('avendesora --stdout scc 0'.split())
    assert result == b'attendant oppress dimple\n'

def test_scc_q1():
    result = subprocess.check_output('avendesora --stdout scc questions.1'.split())
    assert result == b'student camera bleach\n'

def test_scc_q2():
    result = subprocess.check_output('avendesora --stdout scc questions[2]'.split())
    assert result == b'workforce simulcast smelt\n'

def test_login():
    result = subprocess.check_output('avendesora --stdout login'.split())
    assert result == b'racket genetic outbreak earlobe waterway fatality\n'



#
# INTERFACE TO GNUPG PACKAGE
#
# Package for reading and writing text files that may or may not be encrypted.
# File will be encrypted if file path ends in a GPG extension.

from .config import get_setting, override_setting
from inform import debug, display, fatal, is_collection
from shlib import to_path
import gnupg
import io
GPG_EXTENSIONS = ['.gpg', '.asc']


class GnuPG:
    __shared_state = {}

    def __init__(self,
        gpg_id=None, gpg_path=None, gpg_home=None, armor=None
    ):
        if not gpg_id:
            gpg_id = get_setting('gpg_id')
        if not gpg_id:
            gpg_id = self._guess_id()
        self.gpg_id = gpg_id
        override_setting('gpg_id', gpg_id)

        self.gpg_path = to_path(
            gpg_path if gpg_path else get_setting('gpg_executable')
        )
        override_setting('gpg_path', self.gpg_path)

        self.gpg_home = to_path(
            gpg_home if gpg_home else get_setting('gpg_home')
        )
        override_setting('gpg_home', self.gpg_home)

        self.armor = armor if armor is not None else get_setting('gpg_armor')
        override_setting('gpg_armor', self.armor)

        gpg_args = {}
        if self.gpg_path:
            gpg_args.update({'gpgbinary': str(self.gpg_path)})
        if self.gpg_home:
            gpg_args.update({'gnupghome': str(self.gpg_home)})
        self.gpg = gnupg.GPG(**gpg_args)

    def _guess_id(self):
        import socket, getpass
        username = getpass.getuser()
        hostname = socket.gethostname().split('.')
        if len(hostname) <= 2:
            hostname = '.'.join(hostname)
        else:
            # strip off name of local machine
            hostname = '.'.join(hostname[1:])
        return username + '@' + hostname

    def save(self, path, contents):
        if path.suffix.lower() in GPG_EXTENSIONS:
            encrypted = self.gpg.encrypt(contents, self.gpg_id, armor=self.armor)
            if not encrypted.ok:
                fatal('unable to encrypt.', encrypted.stderr, culprit=path, sep='\n')
            else:
                if self.armor:
                    path.write_text(str(encrypted))
                else:
                    path.write_bytes(encrypted)
                path.chmod(0o600)
        else:
            path.write_text(contents)

    def read(self, path):
        # file is only assumed to be encrypted if path has gpg extension
        if path.suffix.lower() in GPG_EXTENSIONS:
            with path.open('rb') as f:
                decrypted = self.gpg.decrypt_file(f)
                if not decrypted.ok:
                    fatal('unable to decrypt.', decrypted.stderr, culprit=path, sep='\n')
            return decrypted.data
        else:
            return path.read_text()

    def open(self, path):
        # file will only be encrypted if path has gpg extension
        self.path = path
        self.stream = io.StringIO()
        return self.stream

    def close(self):
        contents = self.stream.getvalue()
        self.save(self.path, contents)



# File
# Read or write possibly encrypted python code files

# License {{{1
# Copyright (C) 2016 Kenneth S. Kundert
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see http://www.gnu.org/licenses/.


# Imports {{{1
from .preferences import (
    SETTINGS_DIR, DEFAULT_ACCOUNTS_FILENAME, DEFAULT_LOG_FILENAME, 
    DEFAULT_ARCHIVE_FILENAME
)
from shlib import to_path
from inform import debug, display, Error, fatal, narrate, os_error

# File class {{{1
class File:
    def __init__(self, path, gpg=None, contents=None):
        self.path = to_path(path)
        self.gpg = gpg

    def read(self):
        path = self.path
        self.encrypted = path.suffix in ['.gpg', '.asc']
        try:
            if self.encrypted:
                # file is encrypted, decrypt it
                code = self.gpg.read(to_path(path))
            else:
                # file is not encrypted
                code = path.read_text()
        except OSError as err:
            raise Error(os_error(err))

        contents = {}
        compiled = compile(code, str(path), 'exec')
        exec(compiled, contents)
        self.contents = contents
        return contents

    def create(self, contents):
        path = self.path
        try:
            to_path(SETTINGS_DIR).mkdir(parents=True, exist_ok=True)
            if path.exists():
                # file creation (init) requested, but file already exists
                # don't overwrite the file, instead read it so the information 
                # can be used to create any remaining files.
                display("%s: already exists." % path)
                return
            # create the file
            display('%s: creating.' % path)
            if path.suffix in ['.gpg', '.asc']:
                narrate('encrypting.', culprit=path)
                # encrypt it
                self.gpg.save(to_path(path), contents)
            else:
                narrate('not encrypting.', culprit=path)
                # file is not encrypted
                with path.open('w') as f:
                    f.write(contents)
        except OSError as err:
            raise Error(os_error(err))




# Password Generator

# License {{{1
# Copyright (C) 2016 Kenneth S. Kundert
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see http://www.gnu.org/licenses/.


# Imports {{{1
from .account import Account
from .config import get_setting
from .dialog import show_list_dialog
from .files import File
from .gpg import GnuPG
from .preferences import (
    ACCOUNTS_FILE_INITIAL_CONTENTS, CONFIG_FILE_INITIAL_CONTENTS,
    CONFIG_FILENAME, DEFAULT_ACCOUNTS_FILENAME, DEFAULT_TEMPLATES_FILENAME,
    HASHES_FILENAME, HASH_FILE_INITIAL_CONTENTS, SETTINGS_DIR,
    TEMPLATES_FILE_INITIAL_CONTENTS,
)
from .secrets import Hidden
from .title import Title
from .utilities import generate_random_string, validate_componenets
from inform import debug, Error, fatal, terminate, terminate_if_errors, notify
from shlib import to_path
from urllib.parse import urlparse

# PasswordGenerator class{{{1
class PasswordGenerator:
    def __init__(self, gpg_id=None, init=False):
        validate_componenets()
        self.gpg = GnuPG(gpg_id)
        if init:
            self.initialize()
            terminate()

        # Now open any accounts files found
        for filename in get_setting('accounts_files', []):
            try:
                path = to_path(SETTINGS_DIR, filename)
                account_file = File(path, self.gpg)
                contents = account_file.read()
                if 'master_password' in contents:
                    self.add_master_to_accounts(contents['master_password'])
                self.add_file_info_to_accounts(account_file)
            except Error as err:
                err.terminate()
        terminate_if_errors()

    def initialize(self):
        def split(s, l=72):
            # Break long string into a series of adjacent shorter strings
            if len(s) < l:
                return '"%s"' % s
            chunks = ['    "%s"' % s[i:i+l] for i in range(0, len(s), l)]
            return '\n' + '\n'.join(chunks) + '\n'

        def dict_to_str(d):
            lines = ['{']
            for k in sorted(d):
                lines.append("    '%s': '%s'," % (k, d[k]))
            lines.append('}')
            return '\n'.join(lines)

        fields = {
            'dict_hash': repr(get_setting('dict_hash')),
            'secrets_hash': repr(get_setting('secrets_hash')),
            'charsets_hash': repr(get_setting('charsets_hash')),
            'accounts_file': repr(DEFAULT_ACCOUNTS_FILENAME),
            'templates_file': repr(DEFAULT_TEMPLATES_FILENAME),
            'log_file': repr(get_setting('log_file')),
            'archive_file': repr(get_setting('archive_file')),
            'gpg_id': repr(self.gpg.gpg_id),
            'gpg_home': repr(get_setting('gpg_home')),
            'gpg_armor': repr(get_setting('gpg_armor')),
            'gpg_executable': repr(get_setting('gpg_executable')),
            'default_field': repr(get_setting('default_field')),
            'default_vector_field': repr(get_setting('default_vector_field')),
            'display_time': repr(get_setting('display_time')),
            'browsers': dict_to_str(get_setting('browsers')),
            'default_browser': repr(get_setting('default_browser')),
            'required_protocols': repr(get_setting('required_protocols')),
            'xdotool_executable': repr(get_setting('xdotool_executable')),
            'xsel_executable': repr(get_setting('xsel_executable')),
            'section': '{''{''{''1',
            'master_password': split(Hidden.hide(generate_random_string(72))),
        }

        # create the config file
        path = to_path(SETTINGS_DIR, CONFIG_FILENAME)
        f = File(path, self.gpg)
        f.create(CONFIG_FILE_INITIAL_CONTENTS.format(**fields))

        # create the hashes file
        path = to_path(SETTINGS_DIR, HASHES_FILENAME.format(**fields))
        f = File(path, self.gpg)
        f.create(HASH_FILE_INITIAL_CONTENTS.format(**fields))

        # create the initial accounts file
        path = to_path(SETTINGS_DIR, DEFAULT_ACCOUNTS_FILENAME)
        f = File(path, self.gpg)
        f.create(ACCOUNTS_FILE_INITIAL_CONTENTS.format(**fields))

        # create the templates file
        path = to_path(SETTINGS_DIR, DEFAULT_TEMPLATES_FILENAME)
        f = File(path, self.gpg)
        f.create(TEMPLATES_FILE_INITIAL_CONTENTS.format(**fields))

    def get_account(self, name):
        if not name:
            raise Error('no account specified.')
        for account in Account.all_accounts():
            if account.matches_exactly(name):
                account.initialize()
                return account
        raise Error('not found.', culprit=name)

    def find_accounts(self, target):
        accounts = []
        for account in Account.all_accounts():
            if account.id_contains(target):
                accounts.append(account)
        return accounts

    def search_accounts(self, target):
        accounts = []
        for account in Account.all_accounts():
            if account.account_contains(target):
                accounts.append(account)
        return accounts

    def discover_account(self):
        # get and parse the title
        data = Title().get_data()

        # split the url into basic components if found
        url = data.get('url')
        if url:
            url = urlparse(url)
            data['protocol'] = url.scheme
            data['host'] = url.netloc
            data['path'] = url.path

        # sweep through accounts to see if any recognize this title data
        # recognizer may fund the following fields in data:
        #     rawdata: the original title
        #     title: the processed title
        #     url: the full url
        #     browser: the name of the browser
        #     protocol: the url scheme (ex. http, https, ...)
        #     host: the url host name or IP address
        #     path: the path component of the url
        #           does not include options or anchor
        matches = {}
        for account in Account.all_accounts():
            secret = account.recognize(data)
            if secret:
                matches[account.get_name()] = secret

        if not matches:
            msg = 'cannot find appropriate account.'
            notify(msg)
            raise Error(msg)
        if len(matches) > 1:
            choice = show_list_dialog(sorted(matches.keys()))
            return choice, matches[choice]
        else:
            return matches.popitem()

    def add_master_to_accounts(self, master):
        for account in Account.all_accounts():
            if not hasattr(account, 'master'):
                account.master = master

    def add_file_info_to_accounts(self, file_info):
        for account in Account.all_accounts():
            if not hasattr(account, '_file_info'):
                account._file_info = file_info



# Avendesora Password Generator Preferences
#
# Copyright (C) 2016 Kenneth S. Kundert

# License {{{1
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see http://www.gnu.org/licenses/.

# Imports {{{1
from .files import File
from .preferences import (
    BROWSERS, CHARSETS_MD5, CONFIG_FILENAME, DEFAULT_ARCHIVE_FILENAME,
    DEFAULT_BROWSER, DEFAULT_DISPLAY_TIME, DEFAULT_FIELD, DEFAULT_LOG_FILENAME,
    DEFAULT_VECTOR_FIELD, DICTIONARY_MD5, GPG_ARMOR, GPG_EXECUTABLE, GPG_HOME,
    REQUIRED_PROTOCOLS, SECRETS_MD5, SETTINGS_DIR, XDOTOOL_EXECUTABLE,
    XSEL_EXECUTABLE
)
from shlib import to_path
from inform import Error, warn
from textwrap import dedent
import re
from appdirs import user_config_dir

# Globals {{{1
Config = {}

# read_config() {{{1
def read_config():
    # First open the config file
    try:
        config_file = File(to_path(SETTINGS_DIR, CONFIG_FILENAME))
        contents = config_file.read()
        Config.update({k.lower(): v for k,v in contents.items()})
    except Error as err:
        #warn(err) # log() might be better, but log file is not available yet
        pass # config file is optional

# get_setting() {{{1
def get_setting(name, default=None):
    name = name.lower()
    try:
        return Config[name]
    except KeyError:
        if name == 'log_file':
            return DEFAULT_LOG_FILENAME
        if name == 'archive_file':
            return DEFAULT_ARCHIVE_FILENAME
        if name == 'default_field':
            return DEFAULT_FIELD
        if name == 'default_vector_field':
            return DEFAULT_VECTOR_FIELD
        if name == 'display_time':
            return DEFAULT_DISPLAY_TIME
        if name == 'xdotool_executable':
            return XDOTOOL_EXECUTABLE
        if name == 'xsel_executable':
            return XSEL_EXECUTABLE
        if name == 'gpg_executable':
            return GPG_EXECUTABLE
        if name == 'gpg_home':
            return GPG_HOME
        if name == 'gpg_armor':
            return GPG_ARMOR
        if name == 'browsers':
            return BROWSERS
        if name == 'default_browser':
            return DEFAULT_BROWSER
        if name == 'required_protocols':
            return REQUIRED_PROTOCOLS
        if name == 'dict_hash':
            return DICTIONARY_MD5
        if name == 'secrets_hash':
            return SECRETS_MD5
        if name == 'charsets_hash':
            return CHARSETS_MD5
        return default

# override_setting() {{{1
def override_setting(name, value):
    Config[name.lower()] = value



# Avendesora Password Generator Preferences
#
# Copyright (C) 2016 Kenneth S. Kundert

# License {{{1
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see http://www.gnu.org/licenses/.

# Imports {{{1
from textwrap import dedent
import re
from appdirs import user_config_dir

# Filenames {{{1
SETTINGS_DIR = user_config_dir('avendesora')
DICTIONARY_FILENAME = 'words'
CONFIG_FILENAME = 'config'
    # config file must be unencrypted (it contains the gpg settings)
HASHES_FILENAME = 'hashes'
DEFAULT_ACCOUNTS_FILENAME = 'accounts.gpg'
DEFAULT_TEMPLATES_FILENAME = 'templates'
DEFAULT_LOG_FILENAME = 'log'
    # log file will be encrypted if you add .gpg or .asc extension
DEFAULT_ARCHIVE_FILENAME = 'archive.gpg'

# Settings {{{1
INDENT = '    '
LABEL_COLOR = 'yellow'
    # choose from normal, black, red, green, yellow, blue, magenta, cyan, white
INITIAL_AUTOTYPE_DELAY = 0.0
DEFAULT_FIELD = 'passcode'
DEFAULT_VECTOR_FIELD = 'questions'
DEFAULT_DISPLAY_TIME = 60


# Utility programs {{{1
# use absolute paths for xdotool and xsel so they cannot be maliciously replaced
XDOTOOL_EXECUTABLE = '/usr/bin/xdotool'
XSEL_EXECUTABLE = '/usr/bin/xsel'

# GPG Settings
GPG_EXECUTABLE = '/usr/bin/gpg2'
GPG_HOME = '~/.gnupg'
GPG_ARMOR = True

# Signatures {{{1
# These signatures must be the sha1 signatures for the corresponding files
# Regenerate them with 'sha1sum <filename>'
# These are used in creating the initial master password file.
SECRETS_MD5 = '0779023f5f1687b92b094a2af5de5bd1'
CHARSETS_MD5 = '405332bcb8330b7502d292991026e328'
DICTIONARY_MD5 = '11fe5bc734f4a956c37d7cb3da16ab3f'

# Browsers {{{1
# Associate a command with a browser key.
# The command must contain a single %s, which is replaced with URL.
BROWSERS = {
    'f': 'firefox -new-tab %s',
    'g': 'google-chrome %s',
    't': 'torbrowser %s',
    'x': 'xdg-open %s', # system default browser
}
DEFAULT_BROWSER = 'x'


# Account Recognition {{{1
REQUIRED_PROTOCOLS = ['https']
    # If one of these protocols are specified in the recognition url, then that
    # protocol must be used in the browser

# Initial config file {{{1
CONFIG_FILE_INITIAL_CONTENTS = dedent('''\
    # List of the files that contain account information
    accounts_files = [{accounts_file}, {templates_file}]

    # The desired location of the log file (relative to config directory).
    # Adding a suffix of .gpg or .asc causes the file to be encrypted
    # (otherwise it can leak account names). Use None to disable logging.
    log_file = {log_file}

    # The desired location of the archive file (relative to config director).
    # End the path in .gpg. Use None to disable archiving.
    archive_file = {archive_file}

    # Various settings
    default_field = {default_field}
    default_vector_field = {default_vector_field}
    display_time = {display_time}
    browsers = {browsers}
    default_browser = {default_browser}
    required_protocols = {required_protocols}

    # Information used by GPG when encrypting and decrypting files.
    gpg_id = {gpg_id}
    gpg_armor = {gpg_armor}
    gpg_home = {gpg_home}
    gpg_executable = {gpg_executable}

    # Utilities
    xdotool_executable = {xdotool_executable}
    xsel_executable = {xsel_executable}

    # Hashes
    charsets_hash = '{dict_hash}'
    secrets_hash = '{charsets_hash}'
    dict_hash = '{secrets_hash}'

    # vim: filetype=python sw=4 sts=4 et ai ff=unix :
''')


# Initial hash file {{{1
HASH_FILE_INITIAL_CONTENTS = dedent('''\
    # Changing the contents of the dictionary, secrets, or charsets will change
    # the secrets you generate, thus you do not want to change these files once
    # you have started using the program. These are hashes for the contents of
    # these files at the time that this file was created. The program will
    # complain if the current contents of these files generate a different
    # hash. Only update these hashes if you know what you are doing.

    dict_hash     = '{dict_hash}'  # DO NOT CHANGE THIS LINE
    secrets_hash  = '{secrets_hash}'  # DO NOT CHANGE THIS LINE
    charsets_hash = '{charsets_hash}'  # DO NOT CHANGE THIS LINE
''')


# Initial accounts file {{{1
ACCOUNTS_FILE_INITIAL_CONTENTS = dedent('''\
    # Account information
    #
    # Add information about each of your accounts to the accounts dictionary.
    #
    # You can use the dedent function to strip leading whitespace from
    # multi-line remarks.  You can use the character sets and exclude function
    # to create alphabets for you character-base passwords.
    #
    # Example:
    # To create an alphabet with all characters except tabs use either:
    #     'alphabet': exclude(PRINTABLE, '\\t')
    # or:
    #     'alphabet': ALPHANUMERIC + PUNCTUATION + ' '

    from avendesora import (
        # Basics
        Account, Hidden,

        # Character sets
        exclude, LOWERCASE, UPPERCASE, LETTERS, DIGITS, ALPHANUMERIC,
        HEXDIGITS, PUNCTUATION, WHITESPACE, PRINTABLE, DISTINGUISHABLE,

        # Secrets
        Password, Passphrase, PIN, Question, MixedPassword, BirthDate,

        # Account Discovery
        RecognizeAll, RecognizeAny, RecognizeTitle, RecognizeURL, RecognizeCWD,
        RecognizeHost, RecognizeUser, RecognizeEnvVar,
    )

    master_password = Hidden({master_password})

    # Accounts
    # AAA {section}
    # BBB {section}
    # CCC {section}
    # DDD {section}
    # EEE {section}
    # FFF {section}
    # GGG {section}
    # HHH {section}
    # III {section}
    # JJJ {section}
    # KKK {section}
    # LLL {section}
    # MMM {section}
    # NNN {section}
    # OOO {section}
    # PPP {section}
    # QQQ {section}
    # RRR {section}
    # SSS {section}
    # TTT {section}
    # UUU {section}
    # VVV {section}
    # WWW {section}
    # XXX {section}
    # YYY {section}
    # ZZZ {section}

    # vim: filetype=python sw=4 sts=4 et ai ff=unix :
''')

# Initial templates file {{{1
TEMPLATES_FILE_INITIAL_CONTENTS = dedent('''\
    # Templates
    #
    # Templates are accounts without master passwords and without much in the way of
    # attributes. Use them to generate pass codes of various types for stealth
    # accounts.

    from avendesora import (
        # Character sets
        exclude, LOWERCASE, UPPERCASE, LETTERS, DIGITS, ALPHANUMERIC,
        HEXDIGITS, PUNCTUATION, WHITESPACE, PRINTABLE, DISTINGUISHABLE,

        # Account
        Account,

        # Secrets
        Password, Passphrase, PIN,
    )

    # Templates
    # AAA {section}
    class Anum4(Account):
        passcode = Password(length=4, alphabet=DISTINGUISHABLE)

    class Anum6(Account):
        passcode = Password(length=6, alphabet=DISTINGUISHABLE)

    class Anum8(Account):
        passcode = Password(length=8, alphabet=DISTINGUISHABLE)

    class Anum10(Account):
        passcode = Password(length=10, alphabet=DISTINGUISHABLE)

    class Anum12(Account):
        aliases = ['anum']
        passcode = Password(length=12, alphabet=DISTINGUISHABLE)

    class Anum14(Account):
        passcode = Password(length=14, alphabet=DISTINGUISHABLE)

    class Anum16(Account):
        passcode = Password(length=16, alphabet=DISTINGUISHABLE)

    class Anum18(Account):
        passcode = Password(length=18, alphabet=DISTINGUISHABLE)

    class Anum20(Account):
        passcode = Password(length=20, alphabet=DISTINGUISHABLE)

    # BBB {section}
    # CCC {section}
    class Char4(Account):
        passcode = Password(length=4, alphabet=ALPHANUMERIC+PUNCTUATION)

    class Char6(Account):
        passcode = Password(length=6, alphabet=ALPHANUMERIC+PUNCTUATION)

    class Char8(Account):
        passcode = Password(length=8, alphabet=ALPHANUMERIC+PUNCTUATION)

    class Char10(Account):
        passcode = Password(length=10, alphabet=ALPHANUMERIC+PUNCTUATION)

    class Char12(Account):
        aliases = ['char']
        passcode = Password(length=12, alphabet=ALPHANUMERIC+PUNCTUATION)

    class Char14(Account):
        passcode = Password(length=14, alphabet=ALPHANUMERIC+PUNCTUATION)

    class Char16(Account):
        passcode = Password(length=16, alphabet=ALPHANUMERIC+PUNCTUATION)

    class Char18(Account):
        passcode = Password(length=18, alphabet=ALPHANUMERIC+PUNCTUATION)

    class Char20(Account):
        passcode = Password(length=20, alphabet=ALPHANUMERIC+PUNCTUATION)

    # DDD {section}
    # EEE {section}
    class Extreme(Account):
        passcode = Passphrase(length=64, alphabet=PRINTABLE)

    # FFF {section}
    # GGG {section}
    # HHH {section}
    # III {section}
    # JJJ {section}
    # KKK {section}
    # LLL {section}
    # MMM {section}
    # NNN {section}
    # OOO {section}
    # PPP {section}
    class Pin4(Account):
        aliases = ['pin']
        passcode = PIN(length=4)

    class Pin6(Account):
        passcode = PIN(length=6)

    class Pin8(Account):
        aliases = ['num']
        passcode = PIN(length=8)

    class Pin10(Account):
        passcode = PIN(length=10)

    class Pin12(Account):
        passcode = PIN(length=12)

    # QQQ {section}
    # RRR {section}
    # SSS {section}
    # TTT {section}
    # UUU {section}
    # VVV {section}
    # WWW {section}
    class Word(Account):
        passcode = Passphrase(length=1)

    class Word2(Account):
        aliases = ['pair']
        passcode = Passphrase(length=2)

    class Word4(Account):
        aliases = ['word', 'xkcd']
        passcode = Passphrase(length=4)

    class Word6(Account):
        passcode = Passphrase(length=6)

    class Word8(Account):
        passcode = Passphrase(length=8)

    # XXX {section}
    # YYY {section}
    # ZZZ {section}

    # vim: filetype=python sw=4 sts=4 et ai ff=unix :
''')

# Fields {{{1
# Fields reserved for use by Avendesora
TOOL_FIELDS = ['aliases', 'default', 'master', 'discovery', 'browser']

